from typing import Callable, Any, Optional, Union, Type, TYPE_CHECKING
from functools import update_wrapper


class MethodProxy:
    def __init__(
        self,
        f: Callable,
        method_name: str,
        args: Optional[tuple] = None,
        kwargs: Optional[dict] = None,
    ) -> None:
        self.f = f
        self.method_name = method_name
        self.args = args
        self.kwargs = kwargs
        update_wrapper(self, f)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.f(*args, **kwargs)


def wrapper(
    method_name: str, args: Optional[tuple] = None, kwargs: Optional[dict] = None
) -> Callable[[Callable], MethodProxy]:
    def inner(f: Callable) -> MethodProxy:
        return MethodProxy(f=f, method_name=method_name, args=args, kwargs=kwargs)

    return inner


class DecoratorFactory:
    def __getattr__(self, item: str) -> Callable[[Callable], MethodProxy]:
        def decorator(f: Callable) -> MethodProxy:
            return MethodProxy(f=f, method_name=item)

        return decorator

    if TYPE_CHECKING:
        def before_first_request(self, f: Callable) -> Callable: ...
        def teardown_appcontext(self, f: Callable) -> Callable: ...
        def shell_context_processor(self, f: Callable) -> Callable: ...
        def before_request(self, f: Callable) -> Callable: ...
        def after_request(self, f: Callable) -> Callable: ...
        def teardown_request(self, f: Callable) -> Callable: ...
        def context_processor(self, f: Callable) -> Callable: ...
        def url_value_preprocessor(self, f: Callable) -> Callable: ...
        def url_defaults(self, f: Callable) -> Callable: ...


class FunctionFactory:
    def __getattr__(self, item: str) -> Callable:
        def decorator(*args: Any, **kwargs: Any) -> Callable:
            return wrapper(item, args, kwargs)

        return decorator

    if TYPE_CHECKING:
        def template_filter(self, name: Optional[str] = None) -> Callable[[
            Callable], Callable]: ...

        def template_test(self, name: Optional[str] = None) -> Callable[[
            Callable], Callable]: ...

        def template_global(self, name: Optional[str] = None) -> Callable[[
            Callable], Callable]: ...

        def errorhandler(self, code_or_exception: Union[Type[Exception], int]) -> Callable[[
            Callable], Callable]: ...


df = DecoratorFactory()
ff = FunctionFactory()



import pkgutil
from typing import Iterable, Type, TypeVar, Any
from types import ModuleType
from importlib import import_module

T = TypeVar("T")


def find_modules(pkg_name: str, *, deep: bool = False, lazy: bool = False) -> Iterable[ModuleType]:
    pkg = import_module(pkg_name)
    module_path = getattr(pkg, "__path__", None)
    if module_path is None:
        return [pkg]
    iter_modules_func = pkgutil.walk_packages if deep else pkgutil.iter_modules
    iter_modules = (
        import_module(module_info[1])
        for module_info in iter_modules_func(module_path, pkg.__name__ + ".")
        if not module_info[2]
    )
    return iter_modules if lazy else list(iter_modules)


def get_objs(module: ModuleType, cls: Type[T]) -> Iterable[T]:
    for attr_name in dir(module):
        if not attr_name.startswith("__"):
            attr = getattr(module, attr_name)
            if isinstance(attr, cls):
                yield attr


def import_string(dotted_path: str) -> Any:
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError(
            'Module "{}" does not define a "{}" attribute/class'.format(
                module_path, class_name
            )
        ) from err


def get_objs_in_modules(
    pkg_names: Iterable[str], cls: Type[T], *, deep: bool = False
) -> Iterable[T]:
    obj_ids = set()
    for pkg_name in pkg_names:
        modules = find_modules(pkg_name, deep=deep)
        for module in modules:
            for obj in get_objs(module, cls):
                obj_id = id(obj)
                if obj_id not in obj_ids:
                    obj_ids.add(obj_id)
                    yield obj



from typing import Iterable, Union, Optional

from flask import Flask, Blueprint

from flask_seek.util import get_objs_in_modules
from flask_seek.factory import MethodProxy


def register_blueprints(
    instance: Union[Flask, Blueprint], pkg_names: Iterable[str], *, deep: bool = False
) -> None:
    for obj in get_objs_in_modules(pkg_names, Blueprint, deep=deep):
        instance.register_blueprint(obj)


def register_methods(
    instance: Union[Flask, Blueprint], pkg_names: Iterable[str], *, deep: bool = False
) -> None:
    for obj in get_objs_in_modules(pkg_names, MethodProxy, deep=deep):
        method = getattr(instance, obj.method_name, None)
        if method is not None:
            if obj.args is None:
                method(obj.f)
            else:
                method(*obj.args, **obj.kwargs)(obj.f)


def seek(
    instance: Union[Flask, Blueprint],
    *,
    blueprint_modules: Optional[Iterable[str]] = None,
    blueprint_deep_modules: Optional[Iterable[str]] = None,
    decorator_modules: Optional[Iterable[str]] = None,
    decorator_deep_modules: Optional[Iterable[str]] = None
) -> None:
    if blueprint_modules:
        register_blueprints(instance, blueprint_modules)
    if blueprint_deep_modules:
        register_blueprints(instance, blueprint_deep_modules, deep=True)
    if decorator_modules:
        register_methods(instance, decorator_modules)
    if decorator_deep_modules:
        register_methods(instance, decorator_deep_modules, deep=True)
